import os
import unittest
from unittest import TestResult
import json
import random
from typing import List, Union
from cpanel.cli import configuration
from cpanel.core import JSONType, CPanelEndpoint, endpoint
from cpanel.__main__ import dispatch
from cpanel_api import Api


class TestCore(unittest.TestCase):

	def setUp(self) -> None:
		_, hostname, username, utoken = configuration([], {}, './test/cpanelrc.test')
		self.host: CPanelEndpoint = endpoint(hostname, username, utoken)


	def run(self, result: TestResult) -> None:
		"""Stop after first error."""
		if not result.errors:
			super(TestCore, self).run(result)


	def list_mail_accounts(self) -> List[JSONType]:
		return json.loads(dispatch(self.host, ["list", "mail", "accounts"]))


	def list_mail_filters(self, account: str) -> List[JSONType]:
		return json.loads(dispatch(self.host, ["list", "mail", "filters", account]))


	def get_mail_filter(self, account: str, filtername: str) -> JSONType:
		return json.loads(dispatch(self.host, ["get", "mail", "filter", account, filtername]))


	# Actual tests below.

	def test_list_features(self) -> None:
		features: JSONType = json.loads(dispatch(self.host, ["list", "features"]))
		print(features)

		key: str
		value: int
		for key, value in features.items():
			self.assertTrue(value in (0, 1))


	def test_get_quota(self) -> None:
		quota: JSONType = json.loads(dispatch(self.host, ["get", "quota"]))
		print(quota)

		self.assertTrue(quota['inodes_used'] >= 0)
		self.assertTrue(quota['megabytes_used'] >= 0)


	def test_get_usage(self) -> None:
		usage: List[JSONType] = json.loads(dispatch(self.host, ["get", "usage"]))
		print(usage)

		self.assertTrue(len(usage) > 0)
		item: JSONType
		for item in usage:
			self.assertTrue(len(item['id']) > 0)
			self.assertTrue(int(item['usage']) >= 0)


	def test_get_stats(self) -> None:
		display: List[str] = ["hostname", "machinetype", "cpanelversion"]
		stats: List[JSONType] = json.loads(dispatch(self.host, ["get", "stats"] + display))
		print(stats)

		self.assertTrue(len(stats) > 0)
		stat: JSONType
		for stat in stats:
			self.assertTrue(stat['id'] in display)
			self.assertTrue(len(stat['value']) > 0)


	def test_list_accounts(self) -> None:
		accounts: List[JSONType] = json.loads(dispatch(self.host, ["list", "accounts"]))
		print(accounts)

		self.assertTrue(len(accounts[0]['user']) > 0)


	def test_get_account(self) -> None:
		account: JSONType = json.loads(dispatch(self.host, ["get", "account"]))
		print(account)

		self.assertTrue(len(account['user']) > 0)


	def test_list_subaccounts(self) -> None:
		subaccounts: List[JSONType] = json.loads(dispatch(self.host, ["list", "subaccounts"]))
		print(subaccounts)

		self.assertTrue(len(subaccounts) > 0)
		subaccount: JSONType
		for subaccount in subaccounts:
			self.assertTrue(len(subaccount['guid']) > 0)


	def test_cache(self) -> None:
		update: JSONType = json.loads(dispatch(self.host, ["update", "cache"]))
		print(update)
		self.assertTrue(update['cache_id'] > 0)

		read: JSONType = json.loads(dispatch(self.host, ["read", "cache"]))
		print(read)
		self.assertTrue(read['cache_id'] > 0)
		self.assertEqual(read['cache_id'], update['cache_id'])


	def test_locales(self) -> None:
		original: JSONType = json.loads(dispatch(self.host, ["get", "locale"]))
		print(original)
		self.assertTrue(len(original['locale']) > 0)

		locales: List[JSONType] = json.loads(dispatch(self.host, ["list", "locales"]))
		locale: JSONType = locales[0]
		print(locale)
		self.assertTrue(len(locale['locale']) > 0)

		r: str = dispatch(self.host, ["set", "locale", locale['locale']])
		self.assertEqual(r, "OK")

		temporary: JSONType = json.loads(dispatch(self.host, ["get", "locale"]))
		print(temporary)
		self.assertEqual(temporary['locale'], locale['locale'])

		r: str = dispatch(self.host, ["set", "locale", original['locale']])
		self.assertEqual(r, "OK")


	def test_get_style(self) -> None:
		style: JSONType = json.loads(dispatch(self.host, ["get", "style"]))
		print(style)

		self.assertTrue(style['name'] in ("basic",  "dark", "light", "glass"))


	def test_get_theme(self) -> None:
		themes: List[JSONType] = json.loads(dispatch(self.host, ["list", "themes"]))
		print(themes)

		theme: JSONType = json.loads(dispatch(self.host, ["get", "theme"]))
		print(theme)

		self.assertTrue(theme in themes)


	def test_list_dir_indexing(self) -> None:
		indexing: JSONType = json.loads(dispatch(self.host, ["list", "dir", "indexing", "/.cpanel"]))
		print(indexing)

		self.assertTrue(indexing['current']['state']['index_type'] in ("inherit",  "disabled", "standard", "fancy"))


	def test_set_dir_indexing(self) -> None:
		original: JSONType = json.loads(dispatch(self.host, ["get", "dir", "indexing", "/.cpanel"]))
		print(original)

		self.assertTrue(original in ("inherit",  "disabled", "standard", "fancy"))

		r: str = dispatch(self.host, ["set", "dir", "indexing", "/.cpanel", str(original)])
		self.assertEqual(r, "OK")


	def test_list_dir_privacy(self) -> None:
		privacy: JSONType = json.loads(dispatch(self.host, ["list", "dir", "privacy", "/.cpanel"]))
		print(privacy)

		self.assertTrue(privacy['current']['state']['protected'] in (0, 1))


	def test_enable_disable_dir_privacy(self) -> None:
		original: JSONType = json.loads(dispatch(self.host, ["get", "dir", "privacy", "/.cpanel"]))
		print(original)

		self.assertTrue(original['protected'] in (0, 1))

		r: str = dispatch(self.host, ["enable", "dir", "privacy", "/.cpanel"])
		self.assertEqual(r, "OK")

		r = dispatch(self.host, ["disable", "dir", "privacy", "/.cpanel"])
		self.assertEqual(r, "OK")

		if original['protected'] == 0:
			r = dispatch(self.host, ["disable", "dir", "privacy", "/.cpanel"])
			self.assertEqual(r, "OK")
		else:
			r = dispatch(self.host, ["enable", "dir", "privacy", "/.cpanel"])
			self.assertEqual(r, "OK")


	def test_add_list_delete_dir_user(self) -> None:
		user: str = 'tmp-{}'.format(hex(random.randrange(0, 2 ** 32))[2:])
		print(user)

		r: str = dispatch(self.host, ["add", "dir", "user", "/.cpanel", user, "tiger"])
		self.assertEqual(r, "OK")

		users: List[JSONType] = json.loads(dispatch(self.host, ["list", "dir", "users", "/.cpanel"]))
		print(users)

		created: bool = False
		for u in users:
			created = created or u == user
		self.assertTrue(created)

		r = dispatch(self.host, ["delete", "dir", "user", "/.cpanel", user])
		self.assertEqual(r, "OK")

		users = json.loads(dispatch(self.host, ["list", "dir", "users", "/.cpanel"]))
		print(users)

		deleted: bool = True
		for u in users:
			deleted = deleted and u != user
		self.assertTrue(deleted)


	def test_list_dir_protection(self) -> None:
		protection: JSONType = json.loads(dispatch(self.host, ["list", "dir", "protection", "/.cpanel"]))
		print(protection)

		self.assertTrue(protection['current']['state']['has_leech_protection'] in (0, 1))


	def test_list_mail_accounts(self) -> None:
		emails: List[JSONType] = self.list_mail_accounts()
		print(emails)

		email: JSONType
		for email in emails:
			self.assertTrue(len(email['email']) > 0)


	def test_list_mail_filters(self) -> None:
		emails: List[JSONType] = self.list_mail_accounts()
		if len(emails) == 0: return

		filters: List[JSONType] = self.list_mail_filters(emails[0]['email'])
		print(filters)

		filter: JSONType
		for filter in filters:
			self.assertTrue(len(filter['filtername']) > 0)


	def test_get_mail_filter(self) -> None:
		emails: List[JSONType] = self.list_mail_accounts()
		if len(emails) == 0: return

		filters: List[JSONType] = self.list_mail_filters(emails[0]['email'])
		if len(filters) == 0: return

		filter: JSONType = self.get_mail_filter(emails[0]['email'], filters[0]['filtername'])
		print(filter)

		actions: List[JSONType] = filter['actions']
		action: JSONType
		for action in actions:
			self.assertTrue(len(action['action']) > 0)
			if action['dest'] is not None: self.assertTrue(len(action['dest']) > 0)
			self.assertTrue(int(action['number']) > 0)

		rules: List[JSONType] = filter['rules']
		rule: JSONType
		for rule in rules:
			self.assertTrue(len(rule['match']) > 0)
			self.assertTrue(len(rule['part']) > 0)
			self.assertTrue(len(rule['val']) > 0)
			self.assertTrue(int(rule['number']) > 0)


	def test_set_then_delete_mail_filter(self) -> None:
		emails: List[JSONType] = self.list_mail_accounts()
		if len(emails) == 0: return

		filters: List[JSONType] = self.list_mail_filters(emails[0]['email'])
		if len(filters) == 0: return

		filter: JSONType = self.get_mail_filter(emails[0]['email'], filters[0]['filtername'])
		if len(filter) == 0: return

		name: str = 'tmp-{}'.format(hex(random.randrange(0, 2 ** 32))[2:])
		filter['filtername'] = name
		filterfile: str = './test/{}.json'.format(name)

		with open(filterfile, 'w', encoding = 'utf-8') as stream:
			stream.write(json.dumps(filter, indent = 4, sort_keys = True))

		r: str = dispatch(self.host, ["set", "mail", "filter", emails[0]['email'], filterfile])
		self.assertEqual(r, "OK")

		r = dispatch(self.host, ["delete", "mail", "filter", emails[0]['email'], name])
		self.assertEqual(r, "OK")

		os.remove(filterfile)



import os
import sys
import re
import pkgutil
from configparser import ConfigParser
import logging
from logging import Logger
import cpanel
from typing import List, Mapping, Tuple, Match
from .core import NullableStr, NullableBytes, CPanelError


log: Logger = logging.getLogger(__name__)


def die(message: NullableStr = None) -> None:
	"""Print (optional) message and exit to shell with error."""
	if message:
		print("{}: {}".format(cpanel.__name__, message), file = sys.stderr)
	sys.exit(1)


def eatvalue(args: List[str], shortopt: str, longopt: str) -> Tuple[List[str], NullableStr]:
	"""Scan args list looking for a short or long option with a value.

	args       list of arguments passed to the command line
	shortopt   short option, e.g. -x VALUE
	longopt    long option, e.g. --xxx=VALUE

	Returns tuple:
		args with options and VALUE possibly removed,
		VALUE if option is found, None otherwise
	"""
	reargs: List[str] = []
	value: NullableStr = None
	n: int = len(longopt)

	for i in range(0, len(args)):
		# Eat [ ..., '-x', 'xvalue', ... ]
		if args[i] == shortopt and i + 1 < len(args):
			value = args[i + 1]
		# Eat [ ..., '--xxx=xvalue', ... ]
		elif len(args[i]) > n and args[i][0:n + 1] == longopt + '=':
			value = args[i][n + 1:]
		elif i == 0 or args[i - 1] != shortopt:
			reargs.append(args[i])

	return reargs, value


def eatflag(args: List[str], shortopt: str, longopt: str) -> Tuple[List[str], bool]:
	"""Scan args list looking for a short or long flag option.

	args       list of arguments passed to the command line
	shortopt   short flag option, e.g. -x
	longopt    long flag option, e.g. --xxx

	Returns tuple:
		args with options possibly removed,
		True if option is found, False otherwise
	"""
	reargs: List[str] = []
	flag: bool = False

	for arg in args:
		# Eat [ ..., '-x', ... ] or [ ..., '--xxx', ... ]
		if arg == shortopt or arg == longopt:
			flag = True
		else:
			reargs.append(arg)

	return reargs, flag


def configuration(args: List[str], env: Mapping[str, str], rcfile: str) \
		-> Tuple[List[str], NullableStr, NullableStr, NullableStr]:
	"""Return a tuple with credentials to hit the cPanel API.

	args       list of arguments passed to the command line
	env        dict of current environment variables
	rcfile     an .rc properties file, with name=value lines

	Returns tuple:
		args with option values possibly removed,
		hostname,
		username,
		UAPI token associated to username
	"""
	hostname: NullableStr = None
	username: NullableStr = None
	utoken: NullableStr = None

	# Parse optional CLI values (they override both environment variables and the .rc file).
	# I don’t like argparse, manual parsing is much simpler and readable.
	args, hostname = eatvalue(args, '-H', '--hostname')
	args, username = eatvalue(args, '-U', '--username')
	args, utoken = eatvalue(args, '-T', '--utoken')

	# Check environment variables (they override the .rc file).
	if hostname is None: hostname = env.get('CPANEL_HOSTNAME')
	if username is None: username = env.get('CPANEL_USERNAME')
	if utoken is None: utoken = env.get('CPANEL_UTOKEN')

	# Finally, parse the .rc file.
	if os.path.isfile(rcfile):
		config : ConfigParser = ConfigParser()
		with open(rcfile, 'r') as stream:
			config.read_string('[cpanel]\n' + stream.read())
		try:
			if hostname is None: hostname = config['cpanel']['hostname']
			if username is None: username = config['cpanel']['username']
			if utoken is None: utoken = config['cpanel']['utoken']
		except KeyError as e:
			raise CPanelError("missing property in {}, {}".format(rcfile, str(e)))

	return args, hostname, username, utoken


def version() -> str:
	"""Return package version."""
	return "{} client version {}".format(cpanel.__description__, cpanel.__version__)


def usage(*args: str) -> str:
	"""Return usage string depending on help arguments.

	args arguments after `cpanel help`

	Return multiline string with usage description
	"""

	def is_usage(*args: str) -> bool:
		if len(args) == 0:
			return True
		if len(args) == 1:
			# HACK  Redirect to USAGE to get help for mail and dir submodules
			if args[0][:3].lower() == "mod" or \
					args[0][:4].lower() == "mail" or \
					args[0][:3].lower() == "dir":
				return True
		return False 
		
	# Read USAGE or REFERENCE data file.
	stream: NullableBytes = pkgutil.get_data(__name__, 'USAGE' if is_usage(*args) else 'REFERENCE')

	# HACK  Allow ANSI color codes using a tortuous decoding/encoding chain.
	# See https://stackoverflow.com/questions/14820429/how-do-i-decodestring-escape-in-python-3
	text: NullableStr = stream.decode('unicode-escape').encode('latin-1').decode('utf-8') if stream else None

	if text:
		match: Match[str] | None = None
		if len(args) > 0:
			# Find the appropiate usage section according to help.
			match = re.search(
				r'^Usage:[ \n]+cpanel [A-Za-z\[\]]+ %s' % " ".join(args).rstrip("s"), text, re.M|re.I)

		if match is None:
			# Find the default usage section.
			match = re.search(r'^Usage: cpanel \[OPTIONS\] COMMAND', text, re.M|re.I)

		if match:
			start: int = match.start()
			# ‘---’ marks the end of the section.
			match = re.search(r'---', text[start:], re.M)
			end: int  = start + match.end() if match else len(text)
			return text[start:end].rstrip("-\n")

	return "cpanel: use --help"



__author__ = 'Marco Bravo Mejías'
__email__ = 'marco@layfellow.net'
__version__ = '0.17.2'
__copyright__ = '2022, ' + __author__
__license__ = 'GPLv3'
__description__ = 'A command line interface for the cPanel Unrestricted API.'
__url__ = 'https://github.com/layfellow/cpanel-cli'



import sys
import os
import logging
from logging import Logger
import cpanel
from cpanel_api import Api
from typing import cast, List, Dict, Callable
from .cli import eatflag, configuration, version, die, usage
from .core import NullableStr, CPanelEndpoint, CPanelError, endpoint

HOME: NullableStr = os.environ.get('HOME')
RCFILE: str = (HOME + '/' if HOME else '') + '.' + cpanel.__name__ + 'rc'

#  Switch on debugging information if enviroment variable DEBUG is set to 1.
logging.basicConfig(stream = sys.stderr, level = logging.DEBUG if os.environ.get('DEBUG') else logging.INFO)
log: Logger = logging.getLogger(__name__)

# Force the underlying cpanel_api package to use the same log level as ours.
logging.getLogger('cpanel_api').setLevel(log.getEffectiveLevel())


def dispatch(host: CPanelEndpoint, args: List[str]) -> str:
	"""Make a cPanel API call corresponding to the args.

	host       reference to API endpoint
	args       list of arguments passed to the command line, minus options

	Returns    result data (usually a JSON string)
	"""
	log.debug(str(args))
	cmd: str = " ".join(args[0:3])  # Ignore everything beyond the 3rd arg.
	uapi: Api = host.client.uapi

	def cmd_is(cmd: str, *args: str) -> bool:
		for arg in args:
			if cmd[0:len(arg)] == arg:
				return True
		return False

	r: str = ""

	try:
		if cmd_is(cmd, "list feature"):
			r = host.dump(lambda: uapi.Features.list_features())

		elif cmd_is(cmd, "get quota"):
			r = host.dump(lambda: uapi.Quota.get_quota_info())

		elif cmd_is(cmd, "get usage"):
			r = host.dump(lambda: uapi.ResourceUsage.get_usages())

		elif cmd_is(cmd, "get stat"):
			r = host.dump(lambda: uapi.StatsBar.get_stats(display = "|".join(args[2:])))

		elif cmd_is(cmd, "list account"):
			r = host.dump(lambda: uapi.Resellers.list_accounts())

		elif cmd_is(cmd, "get account"):
			r = host.dump(lambda: uapi.Variables.get_user_information())

		elif cmd_is(cmd, "list subaccount"):
			r = host.dump(lambda: uapi.UserManager.list_users())

		elif cmd_is(cmd, "get subaccount"):
			r = host.dump(lambda: uapi.UserManager.lookup_user(guid = args[2]))

		elif cmd_is(cmd, "create backup"):
			r = host.create_backup(*args[2:])

		elif cmd_is(cmd, "list backup"):
			r = host.dump(lambda: uapi.Backup.list_backups())

		elif cmd_is(cmd, "update cache"):
			r = host.dump(lambda: uapi.CacheBuster.update())

		elif cmd_is(cmd, "read cache"):
			r = host.dump(lambda: uapi.CacheBuster.read())

		elif cmd_is(cmd, "list locale"):
			r = host.dump(lambda: uapi.Locale.list_locales())

		elif cmd_is(cmd, "get locale"):
			r = host.dump(lambda: uapi.Locale.get_attributes())

		elif cmd_is(cmd, "set locale"):
			r = host.check(lambda: uapi.Locale.set_locale(locale = args[2]))

		elif cmd_is(cmd, "list style"):
			r = host.dump(lambda: uapi.Styles.list())

		elif cmd_is(cmd, "get style"):
			r = host.dump(lambda: uapi.Styles.current())

		elif cmd_is(cmd, "set style"):
			r = host.check(lambda: uapi.Styles.update(name = args[2], type = 'default'))

		elif cmd_is(cmd, "default style"):
			r = host.check(lambda: uapi.Styles.set_default(name = args[2], type = 'default'))

		elif cmd_is(cmd, "list theme"):
			r = host.dump(lambda: uapi.Themes.list(show_mail_themes = 1))

		elif cmd_is(cmd, "get theme"):
			r = host.dump(lambda: uapi.Themes.get_theme_base())

		elif cmd_is(cmd, "set theme"):
			r = host.check(lambda: uapi.Themes.update(theme = args[2]))

		elif cmd_is(cmd, "list dir indexing"):
			r = host.dump(lambda: uapi.DirectoryIndexes.list_directories(dir = args[3]))

		elif cmd_is(cmd, "get dir indexing"):
			r = host.dump(lambda: uapi.DirectoryIndexes.get_indexing(dir = args[3]))

		elif cmd_is(cmd, "set dir indexing"):
			r = host.check(lambda: uapi.DirectoryIndexes.set_indexing(dir = args[3], type = args[4]))

		elif cmd_is(cmd, "list dir privacy"):
			r = host.dump(lambda: uapi.DirectoryPrivacy.list_directories(dir = args[3]))

		elif cmd_is(cmd, "get dir privacy"):
			r = host.dump(lambda: uapi.DirectoryPrivacy.is_directory_protected(dir = args[3]))

		elif cmd_is(cmd, "enable dir privacy"):
			r = host.check(lambda: uapi.DirectoryPrivacy.configure_directory_protection(
				dir = args[3], authname = 'protectandserve', enabled = 1))

		elif cmd_is(cmd, "disable dir privacy"):
			r = host.check(lambda: uapi.DirectoryPrivacy.configure_directory_protection(
				dir = args[3], authname = 'protectandserve', enabled = 0))

		elif cmd_is(cmd, "add dir user"):
			r = host.check(lambda: uapi.DirectoryPrivacy.add_user(
				dir = args[3], user = args[4], password = args[5]))

		elif cmd_is(cmd, "delete dir user", "rm dir user", "remove dir user"):
			r = host.check(lambda: uapi.DirectoryPrivacy.delete_user(dir = args[3], user = args[4]))

		elif cmd_is(cmd, "list dir users"):
			r = host.dump(lambda: uapi.DirectoryPrivacy.list_users(dir = args[3]))

		elif cmd_is(cmd, "list dir protection"):
			r = host.dump(lambda: uapi.DirectoryProtection.list_directories(dir = args[3]))

		elif cmd_is(cmd, "check dns"):
			r = host.dump_null(
				lambda: uapi.DNS.ensure_domains_reside_only_locally(domain = args[2]), args[2])

		elif cmd_is(cmd, "authoritative dns"):
			r = host.dump(lambda: uapi.DNS.has_local_authority(domain = args[2]))

		elif cmd_is(cmd, "lookup dns"):
			r = host.dump(lambda: uapi.DNS.lookup(domain = args[2]))

		elif cmd_is(cmd, "list dynamic dns"):
			r = host.dump(lambda: uapi.DynamicDNS.list())

		elif cmd_is(cmd, "create dynamic dns"):
			if len(args) > 4:
				r = host.dump(lambda: uapi.DynamicDNS.create(domain = args[3], description = args[4]))
			else:
				r = host.dump(lambda: uapi.DynamicDNS.create(domain = args[3]))

		elif cmd_is(cmd, "recreate dynamic dns"):
			r = host.check(lambda: uapi.DynamicDNS.recreate(id = args[3]))

		elif cmd_is(cmd, "update dynamic dns"):
			r = host.check(lambda: uapi.DynamicDNS.set_description(id = args[3], description = args[4]))

		elif cmd_is(cmd, "delete dynamic dns", "rm dynamic dns", "remove dynamic dns"):
			r = host.check(lambda: uapi.DynamicDNS.delete(id = args[3]))

		elif cmd_is(cmd, "list domains"):
			r = host.dump(lambda: uapi.DomainInfo.list_domains())

		elif cmd_is(cmd, "list domain data"):
			r = host.dump(lambda: uapi.DomainInfo.domains_data(
				format = 'hash', return_https_redirects_status = 1))

		elif cmd_is(cmd, "get domain data"):
			r = host.dump(lambda: uapi.DomainInfo.single_domain_data(
				domain = args[3], return_https_redirects_status = 1))

		elif cmd_is(cmd, "get domain aliases"):
			r = host.dump(lambda: uapi.DomainInfo.main_domain_builtin_subdomain_aliases())

		elif cmd_is(cmd, "get log settings"):
			r = host.dump(lambda: uapi.LogManager.get_settings())

		elif cmd_is(cmd, "set log settings"):
			r = host.set_log_settings(1, *args[3:])

		elif cmd_is(cmd, "unset log settings"):
			r = host.set_log_settings(0, *args[3:])

		elif cmd_is(cmd, "list log archives"):
			r = host.dump(lambda: uapi.LogManager.list_archives())

		elif cmd_is(cmd, "get bandwidth services"):
			r = host.dump(lambda: uapi.Bandwidth.get_enabled_protocols())

		elif cmd_is(cmd, "get bandwidth retention"):
			r = host.dump(lambda: uapi.Bandwidth.get_retention_periods())

		elif cmd_is(cmd, "list files"):
			r = host.dump(lambda: uapi.Fileman.list_files(
				dirs = args[2] if len(args) > 2 else '/', include_mime = 1, include_permissions = 1, show_hidden = 1))

		elif cmd_is(cmd, "glob files"):
			r = host.dump(lambda: uapi.Fileman.autocompletedir(path = args[2], html = 0))

		elif cmd_is(cmd, "get file info"):
			r = host.dump(lambda: uapi.Fileman.get_file_information(
				path = args[3], include_mime = 1, include_permissions = 1, show_hidden = 1))

		elif cmd_is(cmd, "cat file"):
			sys.stdout.write(host.get_file_contents(args[2]).decode('utf-8'))

		elif cmd_is(cmd, "write file"):
			r = host.write_file(args[2], args[3])

		elif cmd_is(cmd, "upload file"):
			r = host.upload_file(args[2], args[3])

		elif cmd_is(cmd, "delete file trash", "rm file trash", "remove file trash"):
			r = host.check(lambda: uapi.Fileman.empty_trash(older_than = int(args[3]) if len(args) > 3 else 0))

		elif cmd_is(cmd, "count mail account"):
			r = host.dump(lambda: uapi.Email.count_pops())

		elif cmd_is(cmd, "get mail setting"):
			r = host.dump(lambda: uapi.Email.get_client_settings(account = args[3]))

		elif cmd_is(cmd, "list mail account"):
			r = host.dump_extracted('email', lambda: uapi.Email.list_pops())

		elif cmd_is(cmd, "list mail boxes"):
			if(len(args) > 4):
				r = host.dump(lambda: uapi.Email.browse_mailbox(account = args[3], dir = args[4], show_hidden_files = 1))
			if(len(args) > 3):
				r = host.dump(lambda: uapi.Email.browse_mailbox(account = args[3], show_hidden_files = 1))
			else:
				r = host.dump(lambda: uapi.Email.browse_mailbox(show_hidden_files = 1))

		elif cmd_is(cmd, "list mail autoresponder"):
			r = host.dump(lambda: uapi.Email.list_auto_responders(domain = args[3]))

		elif cmd_is(cmd, "count mail autoresponder"):
			r = host.dump(lambda: uapi.Email.count_auto_responders())

		elif cmd_is(cmd, "get mail autoresponder"):
			r = host.dump(lambda: uapi.Email.get_auto_responder(email = args[3]))

		elif cmd_is(cmd, "set mail autoresponder"):
			r = host.set_mail_autoresponder(*args[3:])

		elif cmd_is(cmd, "delete mail autoresponder", "rm mail autoresponder", "remove mail autoresponder"):
			r = host.check(lambda: uapi.Email.delete_auto_responder(email = args[3]))

		elif cmd_is(cmd, "list mail filter"):
			r = host.dump_extracted('filtername', lambda: uapi.Email.list_filters(account = args[3]))

		elif cmd_is(cmd, "get mail filter"):
			r = host.dump(lambda: uapi.Email.get_filter(account = args[3], filtername = args[4]))

		elif cmd_is(cmd, "set mail filter"):
			r = host.set_mail_filter(args[3], args[4])

		elif cmd_is(cmd, "delete mail filter", "rm mail filter", "remove mail filter"):
			r = host.check(lambda: uapi.Email.delete_filter(account = args[3], filtername = args[4]))

		elif cmd_is(cmd, "get webmail setting"):
			if len(args) > 3:
				r = host.dump(lambda: uapi.Email.get_webmail_settings(account = args[3]))
			else:
				r = host.dump(lambda: uapi.Email.get_webmail_settings())

		elif cmd_is(cmd, "list webmail app"):
			r = host.dump(lambda: uapi.WebmailApps.list_webmail_apps())

		else:
			die("unrecognized command, {}".format(cmd))
	except IndexError:
		die("missing arguments for {}".format(cmd))
	except Exception as e:
		# Re-throw exception raised during API call.
		raise e
	return r


def main() -> None:
	hostname: NullableStr
	username: NullableStr
	utoken: NullableStr

	_help: bool
	_version: bool

	try:
		args: List[str] = sys.argv[1:]

		# Parse options (I don’t like argparse, manual parsing is much simpler and readable.)
		args, _help = eatflag(args, '-h', '--help')
		args, _version = eatflag(args, '-V', '--version')

		if _version:
			print(version())
		elif len(args) < 1:
			print(usage())
		elif _help:
			print(usage(*args[1:]))
		elif args[0] == 'version':
			print(version())
		elif args[0] == 'help':
			if len(args) < 2:
				print(usage())
			else:
				print(usage(*args[1:]))
		else:
			args, hostname, username, utoken = configuration(args, os.environ, RCFILE)
			if hostname is None: die("missing cPanel hostname, use cpanel --help")
			if username is None: die("missing cPanel username, use cpanel --help")
			if utoken is None: die("missing cPanel UAPI token, use cpanel --help")
			log.debug("hostname: {}, username: {}, utoken: {}".format(hostname, username, utoken))

			r: str = dispatch(endpoint(hostname, username, utoken), args)
			if len(r) > 0:
				print(r)

	except Exception as e:
		if log.isEnabledFor(logging.DEBUG):
			raise e
		else:
			die(str(e))


if __name__ == '__main__':
	main()



import os
import sys
import json
import logging
import requests
import time
import parsedatetime
from datetime import datetime, timezone, timedelta
from urllib.parse import urljoin
from logging import Logger
from json import JSONDecodeError
import cpanel
from cpanel_api import CPanelApi, Result, ClientError
from typing import cast, Any, Union, Dict, List, Mapping, Callable

NullableStr = Union[str, None]
NullableBytes = Union[bytes, None]
JSONType = Dict[str, Any]

log: Logger = logging.getLogger(__name__)


class CPanelError(Exception):
	pass


class CPanelEndpoint:
	"""A wrapper over CPanelApi with some utility functions."""

	def __init__(self, client: CPanelApi) -> None:
		self.client = client


	def safely(self, r: Result, method: Callable) -> str:
		"""Check for possible errors in API call result, then call method.

		r          API call result
		method     callback to process r if r has no errors

		Returns processed r or empty string
		"""
		if r.status == 1 and r.errors is None:
			return method()
		else:
			raise CPanelError(r.errors[0])
		return ""


	def extract(self, r: Result, key: str) -> str:
		"""Get r['data'] values corresponding to key.

		r          API call result
		key        key name

		Returns stringified JSON array with key: value pairs
		"""
		data: List[Mapping[str, str]] = []
		for datum in r.data:
			data.append({ key: datum[key] })
		return json.dumps(data, indent = 4, sort_keys = True)


	def check(self, apicall: Callable) -> str:
		"""Call API and check if request was OK, do not print results.

		apicall    deferred API call

		Returns "OK" or prints error
		"""
		r: Result = apicall()
		return self.safely(r, lambda: "OK")


	def dump(self, apicall: Callable) -> str:
		"""Call API and get stringified JSON result.

		apicall    deferred API call

		Returns stringified JSON data
		"""
		r: Result = apicall()
		return self.safely(r, lambda: json.dumps(r.data, indent = 4, sort_keys = True))


	def dump_extracted(self, key: str, apicall: Callable) -> str:
		"""Call API and get stringified JSON result filtered by key.

		key        key name to filter result
		apicall    deferred API call

		Returns stringified JSON data
		"""
		r: Result = apicall()
		return self.safely(r, lambda: self.extract(r, key))


	def dump_null(self, apicall: Callable, replace: str) -> str:
		"""Call API and get stringified JSON result, but replaces nulls.

		apicall    deferred API call
		replace    string to replace nulls with

		Returns stringified JSON data
		"""
		r: Result = apicall()
		i = 0
		for item in r['data']:
			if item is None:
				r['data'][i] = replace
			i += 1

		return self.safely(r, lambda: json.dumps(r.data, indent = 4, sort_keys = True))


	def create_backup(self, *args: str) -> str:
		"""Create a backup tarball and store it on a remote server.

		args      variable argument list with username, password, host,
		          optional directory and optional confirmation email

		Returns "OK" or prints error
		"""
		# kwargs for CPanelAPi call to Backup.fullbackup_to_*().
		parameters: Mapping[str, NullableStr] = {}

		try:
			if args[0] == 'ftp' or args[0] == 'scp':
				parameters['username'] = args[1]
				parameters['password'] = args[2]
				parameters['host'] = args[3]
				if len(args) > 4:
					parameters['directory'] = args[4]
				if len(args) > 5:
					parameters['email'] = args[5]
			elif args[0] == 'home':
				if len(args) > 1:
					parameters['email'] = args[1]
			else:
				raise CPanelError("create backup target must be ftp, home or scp")

			log.debug(str(args))

			if args[0] == 'ftp':
				return self.check(lambda: self.client.uapi.Backup.fullbackup_to_ftp(**parameters))
			elif args[0] == 'scp':
				return self.check(lambda: self.client.uapi.Backup.fullbackup_to_scp_with_password(**parameters))

			return self.check(lambda: self.client.uapi.Backup.fullbackup_to_homedir(**parameters))

		except IndexError:
			raise CPanelError("missing arguments for create backup")


	def set_log_settings(self, flag: int, *args: str) -> str:
		"""Enable or disable log archival settings.

		flag        0 to enable, 1 to disable
		args        List of archival settings to enable or disable

		Returns "OK" or prints error
		"""
		parameters: Mapping[str, NullableStr] = {}

		log.debug(str(args))

		if len(args) == 0:
			raise CPanelError("missing arguments for set log settings")
		if 'archive' in args:
			parameters['archive_logs'] = str(flag)
		if 'prune' in args:
			parameters['prune_archive'] = str(flag)
		if len(parameters) == 0:
			raise CPanelError("unrecognized arguments for set log settings")

		return self.check(lambda: self.client.uapi.LogManager.set_settings(**parameters))


	def get_file_contents(self, filepath: str) -> bytes:
		"""Return the contents of remote filepath encoded using UTF-8.

		filepath    path to remote file
		"""
		dirname: str = os.path.dirname(filepath)
		if len(dirname) == 0:
			dirname = "/"
		basename: str = os.path.basename(filepath)

		r: Result = self.client.uapi.Fileman.get_file_content(
			dir = dirname, file = basename, to_charset = 'utf-8')
		if r.status != 1 or r.errors is not None:
			raise CPanelError(r.errors[0])

		return r.data.content.encode('utf-8')


	def write_file(self, filepath: str, content: str) -> str:
		"""Write a remote filepath with provided content.

		filepath    path to remote file
		content     content string encoded using UTF-8

		Returns "OK" or prints error
		"""
		dirname: str = os.path.dirname(filepath)
		if len(dirname) == 0:
			dirname = "/"
		basename: str = os.path.basename(filepath)

		# Unescape backslash-escaped codes.
		content = content.encode('raw_unicode_escape').decode('unicode_escape')

		return self.check(lambda: self.client.uapi.Fileman.save_file_content(
			dir = dirname, file = basename, content = content, fallback = 0))


	def set_mail_autoresponder(self, *args: str) -> str:
		"""Create an autoresponder for an email account.

		args      variable argument list with email, and optional
		          from, subject, body, start time and stop time

		Returns "OK" or prints error
		"""
		log.debug(str(args))

		# kwargs for CPanelAPI call to Email.add_auto_responder().
		parameters: Mapping[str, str|int] = {}

		default_subject: str = "This is an automatic message"
		default_body: str = "I’m currently unavailable."

		try:
			email: str = args[0]
			parameters['email'] = email
			parameters['domain'] = email[email.find("@") + 1:]
			parameters['from'] = len(args) > 1 and args[1] or email
			parameters['subject'] = len(args) > 2 and args[2] or default_subject
			parameters['body'] = len(args) > 3 and args[3] or default_body
			parameters['interval'] = 24
			parameters['is_html'] = 1
		except IndexError:
			raise CPanelError("missing arguments for create mail autoresponder")

		cal: parsedatetime.Calendar = parsedatetime.Calendar()

		s: str
		t: time.struct_time
		parsed: int
		starttime: datetime
		endtime: datetime

		s = len(args) > 4 and args[4] or "now"
		t, parsed = cal.parse(s)
		if parsed > 0:
			log.debug(str(t))
			starttime = datetime(*t[:6],tzinfo = timezone(timedelta(seconds = -time.timezone)))
		else:
			raise CPanelError("error parsing start time")

		s = len(args) > 5 and args[5] or "December 24, 2099 11:59 PM"
		t, parsed = cal.parse(s)
		if parsed > 0:
			log.debug(str(t))
			endtime = datetime(*t[:6], tzinfo = timezone(timedelta(seconds = -time.timezone)))
		else:
			raise CPanelError("error parsing end time")

		parameters['start'] = int(starttime.timestamp())
		parameters['stop'] = int(endtime.timestamp())

		log.debug(str(parameters))
		return self.check(lambda: self.client.uapi.Email.add_auto_responder(**parameters))


	def upload_file(self, directory: str, filename: str) -> str:
		"""Upload a local file to a remote directory.

		directory   remote path to directory
		filename    local file name

		Returns "OK" or prints error
		"""
		if not os.path.isfile(filename):
			raise CPanelError("missing file {}".format(filename))

		url: str = urljoin(self.client.base_url, '/execute/Fileman/upload_files')
		permissions: str = "0{}".format(oct(os.stat(filename).st_mode)[-3:])

		# Upload the actual file contents as multipart-encoded form data.
		with open(filename, 'rb') as stream:
			response: requests.Response = self.client.session.post(
				url,
				{ 'dir': directory,
				  'file': filename,
				  'overwrite': 1,
				  'permissions': permissions },
				files = { filename: stream },
				allow_redirects = False,
				headers = { 'Authorization': self.client.auth },
				timeout = self.client.timeout,
				verify = self.client.verify)

		if response.status_code == 401:
			raise CPanelError("Unauthorized")

		r: Result = None
		try:
			r = response.json(object_hook = Result)
		except ValueError:
			raise CPanelError("Bad response")
		return self.safely(r, lambda: "OK")


	def set_mail_filter(self, account: str, filterfile: str) -> str:
		"""Set filter for email account.

		account     email account identifier, usually in the form 'name@example.com'
		filterfile  JSON file with filter rule definitions

		Returns "OK" or prints error
		"""
		if not os.path.isfile(filterfile):
			raise CPanelError("missing JSON filter file {}".format(filterfile))

		try:
			# Check if filterfile is a valid JSON file.
			with open(filterfile, 'r') as stream:
				filters: JSONType = json.loads(stream.read())
		except (IOError, JSONDecodeError) as e:
			raise CPanelError("error parsing JSON filter file {}, {}".format(filterfile, str(e)))

		# kwargs for CPanelAPi call to Email.store_filter().
		parameters: Mapping[str, NullableStr] = {}

		try:
			parameters['filtername'] = filters['filtername']
			parameters['account'] = account

			action: JSONType
			i: int = 1
			for action in filters['actions']:
				parameters['action' + str(i)] = action['action']
				parameters['dest' + str(i)] = action['dest']
				i += 1

			rule: JSONType
			i: int = 1
			for rule in filters['rules']:
				parameters['match' + str(i)] = rule['match']
				parameters['opt' + str(i)] = rule['opt']
				parameters['part' + str(i)] = rule['part']
				parameters['val' + str(i)] = rule['val']
				i += 1

		except KeyError as e:
			raise CPanelError("missing key in JSON filter file {}, {}".format(filterfile, str(e)))

		log.debug(str(parameters))
		return self.check(lambda: self.client.uapi.Email.store_filter(**parameters))


def endpoint(hostname: NullableStr, username: NullableStr, utoken: NullableStr) -> CPanelEndpoint:
	"""CPanelEndpoint factory."""
	return CPanelEndpoint(CPanelApi(hostname, username, utoken, auth_type = 'utoken'))



import cpanel
from sphinx.locale import get_translation

_ = get_translation('index')

def setup(app):
	app.add_css_file('custom.css')

project = cpanel.__description__
copyright = cpanel.__copyright__
author = cpanel.__author__
release = cpanel.__version__
extensions = []
templates_path = []
exclude_patterns = ['build', 'Thumbs.db', '.DS_Store']
html_static_path = ['_static/']
html_title = cpanel.__description__ + " (v" + cpanel.__version__ + ")"
locale_dirs = ['locale/']
gettext_compact = False
html_sidebars = {
    "**": ['logo-text.html', 'globaltoc.html', 'localtoc.html', 'searchbox.html']
}
html_theme = 'sphinx_material'
html_logo = '_static/cpanel-cli-white.svg'
html_theme_options = {
	# Set the name of the project to appear in the navigation.
	'nav_title': _("cPanel CLI"),

	# Set you GA account ID to enable tracking.
	# 'google_analytics_account': 'UA-XXXXX',

	# Specify a base_url used to generate sitemap.xml.
	# 'base_url': 'https://project.github.io/project',

	# Set the color and the accent color.
	'color_primary': 'salmon',
	'color_accent': 'teal',

	# Set the repo location to get a badge with stats
	'repo_url': cpanel.__url__,
	'repo_name': "cpanel-cli",

	# Visible levels of the global TOC; -1 means unlimited
	'globaltoc_depth': 2,
	# If False, expand all TOC entries
	'globaltoc_collapse': True,
	# If True, show hidden TOC entries
	'globaltoc_includehidden': True,

	'version_dropdown': True
}



from setuptools import setup
import cpanel

with open('README.rst', 'r', encoding = 'utf-8') as stream:
	long_description = stream.read()

setup(
	name = 'cpanel-cli',
	version = cpanel.__version__,
	author = cpanel.__author__,
	author_email = cpanel.__email__,
	description = cpanel.__description__,
	long_description = long_description,
	url = cpanel.__url__,
	project_urls = {
		'Bug Tracker': cpanel.__url__ + '/issues'
	},
	classifiers = [
		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
		'Operating System :: OS Independent'
	],
	include_package_data = True,
	packages = ['cpanel'],
	python_requires = '>=3.8',
	install_requires = [
		'cpanel-api>=0.3.0',
		'parsedatetime>=2.6'
	],
	entry_points = {
		'console_scripts': [
			'cpanel = cpanel.__main__:main'
	]}
)
